#   Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Input pipeline for on-device recommendation model.

Process input TF example and prepare train/test datasets.
"""
import collections
import functools
import os
from typing import Dict, List, Tuple

import tensorflow as tf
from tensorflow_examples.lite.model_maker.third_party.recommendation.ml.configs import input_config_pb2
from tensorflow_examples.lite.model_maker.third_party.recommendation.ml.model import utils


INT_DEFAULT_VALUE = 0
STRING_DEFAULT_VALUE = 'UNK'
FLOAT_DEFAULT_VALUE = 0.0


class FeaturesAndVocabsByName(
    collections.namedtuple(
        'FeaturesAndVocabsByName', ['features_by_name', 'vocabs_by_name'])):
  """Holder for intermediate data in input processing pipeline."""
  __slots__ = ()

  def __new__(cls, features_by_name=None, vocabs_by_name=None):
    return super(FeaturesAndVocabsByName, cls).__new__(cls,
                                                       features_by_name,
                                                       vocabs_by_name)


def _prepare_feature_vocab_table(
    feature: input_config_pb2.Feature,
    vocab_file_dir: str) -> tf.lookup.StaticVocabularyTable:
  """Prepare vocabulary table for the feature if needed.

  Prepare the vocabulary table with the specified vocab_name(vocab file name)
  and vocab file directory. Feature type is required for the feature, as it
  will be used to set up the vocabulary table.

  Currently we assume the vocabulary file is a txt file and each line stores
  single feature value.

  Args:
    feature: A feature config input_config_pb2.Feature proto.
    vocab_file_dir: The directory storing vocabulary files.

  Returns:
    A vocabulary table (tf.lookup.StaticVocabularyTable) for the feature.
  """
  assert feature.HasField('vocab_name')
  assert feature.HasField('feature_type')
  assert feature.feature_type in [input_config_pb2.FeatureType.INT,
                                  input_config_pb2.FeatureType.STRING]
  vocab_path = os.path.join(vocab_file_dir, feature.vocab_name)
  num_oov_buckets = 1
  key_type = tf.string if (
      feature.feature_type == input_config_pb2.FeatureType.STRING) else tf.int64
  return tf.lookup.StaticVocabularyTable(
      tf.lookup.TextFileInitializer(
          vocab_path,
          key_dtype=key_type,
          key_index=tf.lookup.TextFileIndex.WHOLE_LINE,
          value_dtype=tf.int64,
          value_index=tf.lookup.TextFileIndex.LINE_NUMBER,
          delimiter='\t'), num_oov_buckets)


def _get_features_vocabs_for_groups(
    feature_groups: List[input_config_pb2.FeatureGroup],
    vocab_file_dir: str = '') -> FeaturesAndVocabsByName:
  """Get feature and vocabulary dictionaries for feature groups.

  Args:
    feature_groups: A list of feature group configs(
      input_config_pb2.FeatureGroup)s.
    vocab_file_dir: The directory storing vocabulary files.

  Returns:
    A FeaturesAndVocabsByName object containing features and vocabs
    dictionaries keyed feature name for features according to the feature
    group list.
  """
  features_by_name = {}
  vocabs_by_name = {}
  for feature_group in feature_groups:
    for feature in feature_group.features:
      features_by_name[feature.feature_name] = feature
      if vocab_file_dir and feature.HasField('vocab_name'):
        vocabs_by_name[feature.feature_name] = _prepare_feature_vocab_table(
            feature, vocab_file_dir)
  return FeaturesAndVocabsByName(
      features_by_name=features_by_name, vocabs_by_name=vocabs_by_name)


def get_features_and_vocabs_by_name(
    input_config: input_config_pb2.InputConfig,
    vocab_file_dir: str = '') -> FeaturesAndVocabsByName:
  """Get feature and vocabulary dictionaries according to input config.

  Args:
    input_config: The input config input_config_pb2.InputConfig proto.
    vocab_file_dir: The directory storing vocabulary files.

  Returns:
    A FeaturesAndVocabsByName object containing features and vocabs
    dictionaries keyed feature name for all features according to input
    config.
  """
  global_features_and_vocabs_by_name = (
      _get_features_vocabs_for_groups(list(input_config.global_feature_groups),
                                      vocab_file_dir))
  activity_features_and_vocabs_by_name = (
      _get_features_vocabs_for_groups(
          list(input_config.activity_feature_groups), vocab_file_dir))
  features_by_name = {
      **global_features_and_vocabs_by_name.features_by_name,
      **activity_features_and_vocabs_by_name.features_by_name
  }
  vocabs_by_name = {
      **global_features_and_vocabs_by_name.vocabs_by_name,
      **activity_features_and_vocabs_by_name.vocabs_by_name
  }
  if input_config.label_feature.HasField('vocab_name'):
    vocabs_by_name[
        input_config.label_feature.feature_name] = _prepare_feature_vocab_table(
            input_config.label_feature, vocab_file_dir)
  return FeaturesAndVocabsByName(
      features_by_name=features_by_name, vocabs_by_name=vocabs_by_name)


def _get_feature_spec(feature_type: input_config_pb2.FeatureType,
                      feature_length: int) -> tf.io.FixedLenFeature:
  """Get feature spec based on feature type and length.

  Args:
    feature_type: The type of the feature (input_config_pb2.FeatureType).
    feature_length: The length of the feature. The feature is expected to be
      a sequence.

  Returns:
    The feature spec to parse serialized examples.

  Raises:
    An error occurring if the feature type is not supported.
  """
  if feature_type == input_config_pb2.FeatureType.INT:
    dtype = tf.int64
    default_value = INT_DEFAULT_VALUE
  elif feature_type == input_config_pb2.FeatureType.STRING:
    dtype = tf.string
    default_value = STRING_DEFAULT_VALUE
  elif feature_type == input_config_pb2.FeatureType.FLOAT:
    dtype = tf.float32
    default_value = FLOAT_DEFAULT_VALUE
  else:
    raise ValueError('Unsupported feature type {}'.format(feature_type))

  return tf.io.FixedLenFeature(
      shape=[feature_length],
      dtype=dtype,
      default_value=[default_value] * feature_length)


def _get_serving_feature_spec(feature_name: str,
                              feature_type: input_config_pb2.FeatureType,
                              feature_length: int) -> tf.TensorSpec:
  if feature_type == input_config_pb2.FeatureType.INT or feature_type == input_config_pb2.FeatureType.STRING:
    dtype = tf.dtypes.int32
  elif feature_type == input_config_pb2.FeatureType.FLOAT:
    dtype = tf.dtypes.float32
  else:
    raise ValueError('Unsupported feature type {}'.format(feature_type))
  return tf.TensorSpec(shape=[feature_length], dtype=dtype, name=feature_name)


def get_serving_input_specs(
    input_config: input_config_pb2.InputConfig) -> Dict[str, tf.TensorSpec]:
  features_by_name = get_features_and_vocabs_by_name(
      input_config).features_by_name
  input_specs = collections.OrderedDict()
  for feature_name, feature in sorted(features_by_name.items()):
    input_specs[feature_name] = _get_serving_feature_spec(
        feature_name, feature.feature_type, feature.feature_length)
  return input_specs


def decode_example(
    serialized_proto: str, features_and_vocabs_by_name: FeaturesAndVocabsByName,
    label_feature_name: str) -> Tuple[Dict[str, tf.Tensor], tf.Tensor]:
  """Decode single serialized example.

  Decode single serialized example, accoring to specified features in input
  config. Perform vocabulary lookup if vocabulary is specified.

  Args:
    serialized_proto: The serialized proto that needs to be decoded.
    features_and_vocabs_by_name: A FeaturesAndVocabsByName object containing
      features and vocabs dictionaries by feature names.
    label_feature_name: Name of the label feature.

  Returns:
    features: The decoded features dictionary.
    label_feature: The label feature.

  """
  features_by_name = features_and_vocabs_by_name.features_by_name
  vocabs_by_name = features_and_vocabs_by_name.vocabs_by_name
  name_to_features = {}
  name_to_features[label_feature_name] = tf.io.FixedLenFeature([1], tf.int64)
  for feature_name, feature in features_by_name.items():
    name_to_features[feature_name] = _get_feature_spec(
        feature_type=feature.feature_type,
        feature_length=feature.feature_length)
  record_features = tf.io.parse_single_example(serialized_proto,
                                               name_to_features)

  features = {}
  for feature_name, feature_value in record_features.items():
    if feature_name in vocabs_by_name:
      features[feature_name] = vocabs_by_name[feature_name].lookup(
          feature_value)
    else:
      features[feature_name] = feature_value
  features = {
      k: v if v.dtype != tf.int64 else tf.cast(v, tf.int32)
      for k, v in features.items()
  }
  label_feature = features[label_feature_name]
  return features, label_feature


def get_input_dataset(data_filepattern: str,
                      input_config: input_config_pb2.InputConfig,
                      vocab_file_dir: str,
                      batch_size: int) -> tf.data.Dataset:
  """An input_fn to create input datasets.

  Args:
    data_filepattern: The file pattern of the input data.
    input_config: The input config input_config_pb2.InputConfig proto.
    vocab_file_dir: The path to the directory storing the vocabulary files.
    batch_size: Batch size of to-be generated dataset.

  Returns:
    A Dataset where each element is a batch of feature dicts.
  """
  features_and_vocabs_by_name = get_features_and_vocabs_by_name(
      input_config, vocab_file_dir)
  if not input_config.HasField('label_feature'):
    raise ValueError('Field label_feature is required.')
  input_files = utils.GetShardFilenames(data_filepattern)
  d = tf.data.TFRecordDataset(input_files)
  d = d.shuffle(len(input_files))
  d = d.repeat()
  d = d.shuffle(buffer_size=10000)
  d = d.map(
      functools.partial(
          decode_example,
          features_and_vocabs_by_name=features_and_vocabs_by_name,
          label_feature_name=input_config.label_feature.feature_name),
      num_parallel_calls=8)
  d = d.batch(batch_size, drop_remainder=True)
  d = d.prefetch(1)
  return d



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""ImageNet preprocessing."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow.compat.v2 as tf


IMAGE_SIZE = 224
CROP_PADDING = 32


class Preprocessor(object):
  """Preprocessing for image classification."""

  def __init__(self,
               input_shape,
               num_classes,
               mean_rgb,
               stddev_rgb,
               use_augmentation=False):
    self.input_shape = input_shape
    self.num_classes = num_classes
    self.mean_rgb = mean_rgb
    self.stddev_rgb = stddev_rgb
    self.use_augmentation = use_augmentation

  def __call__(self, image, label, is_training=True):
    if self.use_augmentation:
      return self._preprocess_with_augmentation(image, label, is_training)
    return self._preprocess_without_augmentation(image, label)

  def _preprocess_with_augmentation(self, image, label, is_training):
    """Image preprocessing method with data augmentation."""
    image_size = self.input_shape[0]
    if is_training:
      image = preprocess_for_train(image, image_size)
    else:
      image = preprocess_for_eval(image, image_size)

    image -= tf.constant(self.mean_rgb, shape=[1, 1, 3], dtype=image.dtype)
    image /= tf.constant(self.stddev_rgb, shape=[1, 1, 3], dtype=image.dtype)

    label = tf.one_hot(label, depth=self.num_classes)
    return image, label

  # TODO(yuqili): Changes to preprocess to support batch input.
  def _preprocess_without_augmentation(self, image, label):
    """Image preprocessing method without data augmentation."""
    image = tf.cast(image, tf.float32)

    image -= tf.constant(self.mean_rgb, shape=[1, 1, 3], dtype=image.dtype)
    image /= tf.constant(self.stddev_rgb, shape=[1, 1, 3], dtype=image.dtype)

    image = tf.compat.v1.image.resize(image, self.input_shape)
    label = tf.one_hot(label, depth=self.num_classes)
    return image, label


def distorted_bounding_box_crop(image,
                                bbox,
                                min_object_covered=0.1,
                                aspect_ratio_range=(0.75, 1.33),
                                area_range=(0.05, 1.0),
                                max_attempts=100):
  """Generates cropped_image using one of the bboxes randomly distorted.

  See `tf.image.sample_distorted_bounding_box` for more documentation.

  Args:
    image: 4-D Tensor of shape [batch, height, width, channels] or 3-D Tensor
      of shape [height, width, channels].
    bbox: `Tensor` of bounding boxes arranged `[1, num_boxes, coords]` where
      each coordinate is [0, 1) and the coordinates are arranged as `[ymin,
      xmin, ymax, xmax]`. If num_boxes is 0 then use the whole image.
    min_object_covered: An optional `float`. Defaults to `0.1`. The cropped area
      of the image must contain at least this fraction of any bounding box
      supplied.
    aspect_ratio_range: An optional list of `float`s. The cropped area of the
      image must have an aspect ratio = width / height within this range.
    area_range: An optional list of `float`s. The cropped area of the image must
      contain a fraction of the supplied image within in this range.
    max_attempts: An optional `int`. Number of attempts at generating a cropped
      region of the image of the specified constraints. After `max_attempts`
      failures, return the entire image.

  Returns:
    cropped image `Tensor`
  """
  with tf.name_scope('distorted_bounding_box_crop'):
    shape = tf.shape(image)
    sample_distorted_bounding_box = tf.image.sample_distorted_bounding_box(
        shape,
        bounding_boxes=bbox,
        min_object_covered=min_object_covered,
        aspect_ratio_range=aspect_ratio_range,
        area_range=area_range,
        max_attempts=max_attempts,
        use_image_if_no_bounding_boxes=True)
    bbox_begin, bbox_size, _ = sample_distorted_bounding_box

    # Crop the image to the specified bounding box.
    offset_y, offset_x, _ = tf.unstack(bbox_begin)
    target_height, target_width, _ = tf.unstack(bbox_size)
    image = tf.image.crop_to_bounding_box(image, offset_y, offset_x,
                                          target_height, target_width)

    return image


def _at_least_x_are_equal(a, b, x):
  """At least `x` of `a` and `b` `Tensors` are equal."""
  match = tf.equal(a, b)
  match = tf.cast(match, tf.int32)
  return tf.greater_equal(tf.reduce_sum(match), x)


def _resize_image(image, image_size, method=None):
  if method is not None:
    tf.compat.v1.logging.info('Use customized resize method {}'.format(method))
    return tf.compat.v1.image.resize([image], [image_size, image_size],
                                     method)[0]
  tf.compat.v1.logging.info('Use default resize_bicubic.')
  return tf.compat.v1.image.resize_bicubic([image], [image_size, image_size])[0]


def _decode_and_random_crop(original_image, image_size, resize_method=None):
  """Make a random crop of image_size."""
  bbox = tf.constant([0.0, 0.0, 1.0, 1.0], dtype=tf.float32, shape=[1, 1, 4])
  image = distorted_bounding_box_crop(
      original_image,
      bbox,
      min_object_covered=0.1,
      aspect_ratio_range=(3. / 4, 4. / 3.),
      area_range=(0.08, 1.0),
      max_attempts=10)
  original_shape = tf.shape(original_image)
  bad = _at_least_x_are_equal(original_shape, tf.shape(image), 3)

  image = tf.cond(bad,
                  lambda: _decode_and_center_crop(original_image, image_size),
                  lambda: _resize_image(image, image_size, resize_method))

  return image


def _decode_and_center_crop(image, image_size, resize_method=None):
  """Crops to center of image with padding then scales image_size."""
  shape = tf.shape(image)
  image_height = shape[0]
  image_width = shape[1]

  padded_center_crop_size = tf.cast(
      ((image_size / (image_size + CROP_PADDING)) *
       tf.cast(tf.minimum(image_height, image_width), tf.float32)), tf.int32)

  offset_height = ((image_height - padded_center_crop_size) + 1) // 2
  offset_width = ((image_width - padded_center_crop_size) + 1) // 2
  image = tf.image.crop_to_bounding_box(image, offset_height, offset_width,
                                        padded_center_crop_size,
                                        padded_center_crop_size)
  image = _resize_image(image, image_size, resize_method)
  return image


def _flip(image):
  """Random horizontal image flip."""
  image = tf.image.random_flip_left_right(image)
  return image


def preprocess_for_train(image,
                         image_size=IMAGE_SIZE,
                         resize_method=tf.image.ResizeMethod.BILINEAR):
  """Preprocesses the given image for evaluation.

  Args:
    image: 4-D Tensor of shape [batch, height, width, channels] or 3-D Tensor
      of shape [height, width, channels].
    image_size: image size.
    resize_method: resize method. If none, use bicubic.

  Returns:
    A preprocessed image `Tensor`.
  """
  image = _decode_and_random_crop(image, image_size, resize_method)
  image = _flip(image)
  image = tf.reshape(image, [image_size, image_size, 3])

  image = tf.image.convert_image_dtype(image, dtype=tf.float32)

  return image


def preprocess_for_eval(image,
                        image_size=IMAGE_SIZE,
                        resize_method=tf.image.ResizeMethod.BILINEAR):
  """Preprocesses the given image for evaluation.

  Args:
    image: 4-D Tensor of shape [batch, height, width, channels] or 3-D Tensor
      of shape [height, width, channels].
    image_size: image size.
    resize_method: if None, use bicubic.

  Returns:
    A preprocessed image `Tensor`.
  """
  image = _decode_and_center_crop(image, image_size, resize_method)
  image = tf.reshape(image, [image_size, image_size, 3])
  image = tf.image.convert_image_dtype(image, dtype=tf.float32)
  return image



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=g-bad-import-order,redefined-builtin
"""APIs for recommendation specifications.

Example:
```python
input_spec = recommendation.spec.InputSpec(
    activity_feature_groups=[
        # Group #1: defines how features are grouped in the first Group.
        dict(
            features=[
                # First feature.
                dict(
                    feature_name='context_movie_id',  # Feature name
                    feature_type='INT',  # Feature type
                    vocab_size=3953,     # ID size (number of IDs)
                    embedding_dim=8,     # Projected feature embedding dim
                    feature_length=10,   # History length of 10.
                ),
                # Maybe more features...
            ],
            encoder_type='CNN',  # CNN encoder (e.g. CNN, LSTM, BOW)
        ),
        # Maybe more groups...
    ],
    label_feature=dict(
        feature_name='label_movie_id',  # Label feature name
        feature_type='INT',  # Label type
        vocab_size=3953,   # Label size (number of classes)
        embedding_dim=8,   # label embedding demension
        feature_length=1,  # Exactly 1 label
    ),
)

model_hparams = recommendation.spec.ModelHParams(
    hidden_layer_dims=[32, 32],  # Hidden layers dimension.
    eval_top_k=[1, 5],           # Eval top 1 and top 5.
    conv_num_filter_ratios=[2, 4],  # For CNN encoder, conv filter mutipler.
    conv_kernel_size=16,            # For CNN encoder, base kernel size.
    lstm_num_units=16,              # For LSTM/RNN, num units.
    num_predictions=10,          # Number of output predictions. Select top 10.
)

spec = recommendation.ModelSpec(
    input_spec=input_spec, model_hparams=model_hparams)
# Or:
spec = model_spec.get(
    'recommendation', input_spec=input_spec, model_hparams=model_hparams)
```
"""

from tensorflow_examples.lite.model_maker.core.data_util.recommendation_config import EncoderType
from tensorflow_examples.lite.model_maker.core.data_util.recommendation_config import Feature
from tensorflow_examples.lite.model_maker.core.data_util.recommendation_config import FeatureGroup
from tensorflow_examples.lite.model_maker.core.data_util.recommendation_config import FeatureType
from tensorflow_examples.lite.model_maker.core.data_util.recommendation_config import InputSpec
from tensorflow_examples.lite.model_maker.core.data_util.recommendation_config import ModelHParams



# Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Golden APIs doc."""

DOCS = {}

# pylint: disable=line-too-long
DOCS[''] = """
Public APIs for TFLite Model Maker, a transfer learning library to train custom TFLite models.

You can install the package with

```bash
pip install tflite-model-maker
```

Typical usage of Model Maker is to create a model in a few lines of code, e.g.:

```python
# Load input data specific to an on-device ML app.
data = DataLoader.from_folder('flower_photos/')
train_data, test_data = data.split(0.9)

# Customize the TensorFlow model.
model = image_classifier.create(train_data)

# Evaluate the model.
accuracy = model.evaluate(test_data)

# Export to Tensorflow Lite model and label file in `export_dir`.
model.export(export_dir='/tmp/')
```

For more details, please refer to our guide:
https://www.tensorflow.org/lite/guide/model_maker.
""".lstrip()

DOCS['audio_classifier'] = """APIs to train an audio classification model.

Tutorial:
https://colab.research.google.com/github/googlecodelabs/odml-pathways/blob/main/audio_classification/colab/model_maker_audio_colab.ipynb

Demo code:
https://github.com/tensorflow/examples/blob/master/tensorflow_examples/lite/model_maker/demo/audio_classification_demo.py
"""

DOCS['config'] = 'APIs for the config of TFLite Model Maker.'

DOCS['image_classifier'] = """APIs to train an image classification model.

Task guide:
https://www.tensorflow.org/lite/tutorials/model_maker_image_classification.
"""

DOCS['model_spec'] = 'APIs for the model spec of TFLite Model Maker.'

DOCS['object_detector'] = 'APIs to train an object detection model.'

DOCS['question_answer'] = """
APIs to train a model that can answer questions based on a predefined text.

Task guide:
https://www.tensorflow.org/lite/tutorials/model_maker_question_answer.
""".lstrip()

DOCS['recommendation'] = """APIs to train an on-device recommendation model.

Demo code:
https://github.com/tensorflow/examples/blob/master/tensorflow_examples/lite/model_maker/demo/recommendation_demo.py
"""

DOCS['recommendation.spec'] = """
APIs for recommendation specifications.

Example:
```python
input_spec = recommendation.spec.InputSpec(
    activity_feature_groups=[
        # Group #1: defines how features are grouped in the first Group.
        dict(
            features=[
                # First feature.
                dict(
                    feature_name='context_movie_id',  # Feature name
                    feature_type='INT',  # Feature type
                    vocab_size=3953,     # ID size (number of IDs)
                    embedding_dim=8,     # Projected feature embedding dim
                    feature_length=10,   # History length of 10.
                ),
                # Maybe more features...
            ],
            encoder_type='CNN',  # CNN encoder (e.g. CNN, LSTM, BOW)
        ),
        # Maybe more groups...
    ],
    label_feature=dict(
        feature_name='label_movie_id',  # Label feature name
        feature_type='INT',  # Label type
        vocab_size=3953,   # Label size (number of classes)
        embedding_dim=8,   # label embedding demension
        feature_length=1,  # Exactly 1 label
    ),
)

model_hparams = recommendation.spec.ModelHParams(
    hidden_layer_dims=[32, 32],  # Hidden layers dimension.
    eval_top_k=[1, 5],           # Eval top 1 and top 5.
    conv_num_filter_ratios=[2, 4],  # For CNN encoder, conv filter mutipler.
    conv_kernel_size=16,            # For CNN encoder, base kernel size.
    lstm_num_units=16,              # For LSTM/RNN, num units.
    num_predictions=10,          # Number of output predictions. Select top 10.
)

spec = recommendation.ModelSpec(
    input_spec=input_spec, model_hparams=model_hparams)
# Or:
spec = model_spec.get(
    'recommendation', input_spec=input_spec, model_hparams=model_hparams)
```
""".lstrip()

DOCS['text_classifier'] = """APIs to train a text classification model.

Task guide:
https://www.tensorflow.org/lite/tutorials/model_maker_text_classification.
"""
# pylint: enable=line-too-long



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Audio classification demo code of Model Maker for TFLite.

Example usage:
python audio_classification_demo.py --export_dir=/tmp

Sample output:
Downloading data from
https://storage.googleapis.com/download.tensorflow.org/data/mini_speech_commands.zip
182083584/182082353 [==============================] - 4s 0us/step
182091776/182082353 [==============================] - 4s 0us/step
Dataset has been downloaded to
/usr/local/google/home/wangtz/.keras/datasets/mini_speech_commands
Processing audio files:
8000/8000 [==============================] - 354s 44ms/file
Cached 7178 audio samples.
Training the model
5742/5742 [==============================] - 29s 5ms/step - loss: 3.2289 - acc:
0.8029 - val_loss: 0.6229 - val_acc: 0.9638
Evaluating the model
15/15 [==============================] - 2s 12ms/step - loss: 1.3569 - acc:
0.9270
Test accuracy: 0.927039
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from absl import app
from absl import flags
from absl import logging
import tensorflow as tf
from tflite_model_maker import audio_classifier
from tflite_model_maker import ExportFormat
from tflite_model_maker import model_spec

FLAGS = flags.FLAGS


def _define_flags():
  """Define CLI flags and their default values."""
  flags.DEFINE_string('export_dir', None,
                      'The directory to save exported files.')
  flags.DEFINE_string('data_dir', None, 'The directory to load dataset from.')
  flags.DEFINE_string('spec', 'audio_browser_fft',
                      'Name of the model spec to use.')
  flags.DEFINE_string(
      'dataset', 'mini_speech_command',
      'Which dataset to use. Supports: `mini_speech_command`, '
      '`bird` and `esc50`')
  flags.mark_flag_as_required('export_dir')


def _download_dataset(filename, url, extracted_folder_name, **kwargs):
  """Downloads the dataset, and returns path to the extracted directory."""
  tf.compat.v1.logging.info('Downloading the dataset.')
  # ${HOME}/.keras/datasets/FILENAME
  filepath = tf.keras.utils.get_file(
      filename, url, cache_subdir='datasets', extract=True, **kwargs)
  # ${HOME}/.keras/datasets/EXTRACTED_FOLDER_NAME
  folder_path = os.path.dirname(filepath)
  folder_path = os.path.join(folder_path, extracted_folder_name)
  print(f'Dataset has been downloaded to {folder_path}')
  return folder_path


def download_bird_dataset(**kwargs):
  """Downloads the bird dataset, and returns path to the directory."""
  return _download_dataset(
      'birds_dataset.zip',
      'https://storage.googleapis.com/laurencemoroney-blog.appspot.com/birds_dataset.zip',
      'small_birds_dataset', **kwargs)


def download_speech_commands_dataset(**kwargs):
  """Downloads demo dataset, and returns directory path."""
  return _download_dataset(
      'mini_speech_commands.zip',
      'https://storage.googleapis.com/download.tensorflow.org/data/mini_speech_commands.zip',
      'mini_speech_commands', **kwargs)


def download_esc50_dataset(**kwargs):
  """Downloads ESC50 dataset, and returns directory path."""
  return _download_dataset(
      'esc-50.zip', 'https://github.com/karoldvl/ESC-50/archive/master.zip',
      'ESC-50-master', **kwargs)


def run(spec,
        data_dir,
        dataset_type,
        export_dir,
        epochs=5,
        batch_size=32,
        **kwargs):
  """Runs demo."""
  spec = model_spec.get(spec)

  if dataset_type == 'esc50':
    # Limit to 2 categories to speed up the demo
    categories = ['dog', 'cat']
    train_data = audio_classifier.DataLoader.from_esc50(
        spec, data_dir, folds=[0, 1, 2, 3], categories=categories)
    validation_data = audio_classifier.DataLoader.from_esc50(
        spec, data_dir, folds=[
            4,
        ], categories=categories, cache=True)
    test_data = audio_classifier.DataLoader.from_esc50(
        spec, data_dir, folds=[
            5,
        ], categories=categories, cache=True)
  elif dataset_type == 'bird':
    if isinstance(spec, audio_classifier.YamNetSpec):
      # In some files, two consecutive bird sounds might be too far apart, so
      # increase the window size to have a higher probability of capturing the
      # bird sound.
      spec = audio_classifier.YamNetSpec(
          keep_yamnet_and_custom_heads=True,
          frame_step=3 * audio_classifier.YamNetSpec.EXPECTED_WAVEFORM_LENGTH,
          frame_length=6 * audio_classifier.YamNetSpec.EXPECTED_WAVEFORM_LENGTH)
    else:
      raise ValueError('Bird dataset can only be used with YAMNet model.')
    batch_size = 128
    epochs = 100
    train_data = audio_classifier.DataLoader.from_folder(
        spec, os.path.join(data_dir, 'train'), cache=True)
    train_data, validation_data = train_data.split(0.8)
    test_data = audio_classifier.DataLoader.from_folder(
        spec, os.path.join(data_dir, 'test'), cache=True)

  else:
    data = audio_classifier.DataLoader.from_folder(spec, data_dir, cache=True)
    train_data, rest_data = data.split(0.8)
    validation_data, test_data = rest_data.split(0.5)

  print('\nTraining the model')
  model = audio_classifier.create(
      train_data,
      spec,
      validation_data,
      batch_size=batch_size,
      epochs=epochs,
      **kwargs)

  print('\nEvaluating the model')
  model.evaluate(test_data)

  print('\nConfusion matrix: ')
  print(model.confusion_matrix(test_data))
  print('labels: ', test_data.index_to_label)

  print('\nExporing the Saved model and TFLite model to {}'.format(export_dir))
  model.export(
      export_dir, export_format=(ExportFormat.TFLITE, ExportFormat.SAVED_MODEL))


def main(_):
  logging.set_verbosity(logging.INFO)

  export_dir = os.path.expanduser(FLAGS.export_dir)

  if not FLAGS.data_dir:
    if FLAGS.dataset == 'esc50':
      data_dir = download_esc50_dataset()
    elif FLAGS.dataset == 'bird':
      data_dir = download_bird_dataset()
    elif FLAGS.dataset == 'mini_speech_command':
      data_dir = download_speech_commands_dataset()
    else:
      raise ValueError('Unsupported dataset type: ', FLAGS.dataset)
  else:
    data_dir = os.path.expanduser(FLAGS.data_dir)

  run(FLAGS.spec, data_dir, FLAGS.dataset, export_dir=export_dir)


if __name__ == '__main__':
  _define_flags()
  app.run(main)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Interactive command line tool for tensorflow.js model conversion."""

from __future__ import print_function, unicode_literals

import json
import os
import re
import sys
import tempfile
import traceback

try:
  import PyInquirer
except ImportError:
  sys.exit("""Please install PyInquirer using following command:
              pip install PyInquirer==1.0.3""")

import h5py
import tensorflow.compat.v2 as tf
from tensorflow.core.framework import types_pb2
from tensorflow.python.saved_model import loader_impl
from tensorflowjs.converters import converter
from tensorflowjs.converters import common

# regex for recognizing valid url for TFHub module.
TFHUB_VALID_URL_REGEX = re.compile(
    # http:// or https://
    r'^(http)s?://', re.IGNORECASE)

# prompt style
prompt_style = PyInquirer.style_from_dict({
    PyInquirer.Token.Separator: '#6C6C6C',
    PyInquirer.Token.QuestionMark: '#FF9D00 bold',
    PyInquirer.Token.Selected: '#5F819D',
    PyInquirer.Token.Pointer: '#FF9D00 bold',
    PyInquirer.Token.Instruction: '',  # default
    PyInquirer.Token.Answer: '#5F819D bold',
    PyInquirer.Token.Question: '',
})


def value_in_list(answers, key, values):
  """Determine user's answer for the key is in the value list.
  Args:
    answer: Dict of user's answers to the questions.
    key: question key.
    values: List of values to check from.
  """
  try:
    value = answers[key]
    return value in values
  except KeyError:
    return False


def get_tfjs_model_type(model_file):
  with open(model_file) as f:
    data = json.load(f)
    print("====", data)
    if 'format' in data:
      return data['format']
    else: # Default to layers model
      return common.TFJS_LAYERS_MODEL_FORMAT


def detect_saved_model(input_path):
  if os.path.exists(os.path.join(input_path, 'assets', 'saved_model.json')):
    return common.KERAS_SAVED_MODEL
  saved_model = loader_impl.parse_saved_model(input_path)
  graph_def = saved_model.meta_graphs[0].object_graph_def
  if graph_def.nodes:
    if 'tf_keras' in graph_def.nodes[0].user_object.identifier:
      return common.KERAS_SAVED_MODEL
  return common.TF_SAVED_MODEL


def detect_input_format(input_path):
  """Determine the input format from model's input path or file.
  Args:
    input_path: string of the input model path
  returns:
    string: detected input format
    string: normalized input path
  """
  input_path = input_path.strip()
  detected_input_format = None
  if re.match(TFHUB_VALID_URL_REGEX, input_path):
    detected_input_format = common.TF_HUB_MODEL
  elif os.path.isdir(input_path):
    if (any(fname.lower().endswith('saved_model.pb')
            for fname in os.listdir(input_path))):
      detected_input_format = detect_saved_model(input_path)
    else:
      for fname in os.listdir(input_path):
        fname = fname.lower()
        if fname.endswith('model.json'):
          filename = os.path.join(input_path, fname)
          if get_tfjs_model_type(filename) == common.TFJS_LAYERS_MODEL_FORMAT:
            input_path = os.path.join(input_path, fname)
            detected_input_format = common.TFJS_LAYERS_MODEL
            break
  elif os.path.isfile(input_path):
    if h5py.is_hdf5(input_path):
      detected_input_format = common.KERAS_MODEL
    elif input_path.endswith('saved_model.pb'):
      input_path = os.path.dirname(input_path)
      detected_input_format = detect_saved_model(input_path)
    elif (input_path.endswith('model.json') and
          get_tfjs_model_type(input_path) == common.TFJS_LAYERS_MODEL_FORMAT):
      detected_input_format = common.TFJS_LAYERS_MODEL

  return detected_input_format, input_path


def input_path_message(answers):
  """Determine question for model's input path.
  Args:
    answer: Dict of user's answers to the questions
  """
  answer = answers[common.INPUT_FORMAT]
  message = 'The original path seems to be wrong, '
  if answer == common.KERAS_MODEL:
    return message + 'what is the path of input HDF5 file?'
  elif answer == common.TF_HUB_MODEL:
    return message + ("what is the TFHub module URL? \n"
                      "(i.e. https://tfhub.dev/google/imagenet/"
                      "mobilenet_v1_100_224/classification/1)")
  else:
    return message + 'what is the directory that contains the model?'


def validate_input_path(input_path, input_format):
  """Validate the input path for given input format.
  Args:
    input_path: input path of the model.
    input_format: model format string.
  """
  path = os.path.expanduser(input_path.strip())
  if not path:
    return 'Please enter a valid path'
  if input_format == common.TF_HUB_MODEL:
    if not re.match(TFHUB_VALID_URL_REGEX, path):
      return """This is not an valid URL for TFHub module: %s,
        We expect a URL that starts with http(s)://""" % path
  elif not os.path.exists(path):
    return 'Nonexistent path for the model: %s' % path
  if input_format in (common.KERAS_SAVED_MODEL, common.TF_SAVED_MODEL):
    is_dir = os.path.isdir(path)
    if not is_dir and not path.endswith('saved_model.pb'):
      return 'The path provided is not a directory or pb file: %s' % path
    if (is_dir and
        not any(f.endswith('saved_model.pb') for f in os.listdir(path))):
      return 'Did not find a .pb file inside the directory: %s' % path
    if input_format == common.KERAS_SAVED_MODEL:
      if detect_saved_model(input_path) != common.KERAS_SAVED_MODEL:
        return 'This is a saved model but not a keras saved model: %s' % path

  if input_format == common.TFJS_LAYERS_MODEL:
    is_dir = os.path.isdir(path)
    if not is_dir and not path.endswith('model.json'):
      return 'The path provided is not a directory or json file: %s' % path
    if is_dir and not any(f.endswith('model.json') for f in os.listdir(path)):
      return 'Did not find the model.json file inside the directory: %s' % path
  if input_format == common.KERAS_MODEL:
    if not h5py.is_hdf5(path):
      return 'The path provided is not a keras model file: %s' % path
  return True


def expand_input_path(input_path):
  """Expand the relative input path to absolute path, and add layers model file
  name to the end if input format is `tfjs_layers_model`.
  Args:
    input_path: input path of the model.
  Returns:
    string: return expanded input path.
  """
  input_path = os.path.expanduser(input_path.strip())
  is_dir = os.path.isdir(input_path)
  if is_dir:
    for fname in os.listdir(input_path):
      if fname.endswith('.json'):
        filename = os.path.join(input_path, fname)
        return filename
  return input_path


def output_path_exists(output_path):
  """Check the existence of the output path.
  Args:
    output_path: input path of the model.
  Returns:
    bool: return true when the output directory exists.
  """
  if os.path.exists(output_path):
    return True
  return False


def generate_arguments(params):
  """Generate the tensorflowjs command string for the selected params.
  Args:
    params: user selected parameters for the conversion.
  Returns:
    list: the argument list for converter.
  """
  args = []
  not_param_list = [common.INPUT_PATH, common.OUTPUT_PATH,
                    'overwrite_output_path', 'quantize']
  no_false_param = [common.SPLIT_WEIGHTS_BY_LAYER, common.SKIP_OP_CHECK]
  for key, value in sorted(params.items()):
    if key not in not_param_list and value is not None:
      if key in no_false_param:
        if value is True:
          args.append('--%s' % (key))
      else:
        args.append('--%s=%s' % (key, value))

  args.append(params[common.INPUT_PATH])
  args.append(params[common.OUTPUT_PATH])
  return args


def is_saved_model(input_format):
  """Check if the input path contains saved model.
  Args:
    input_format: input model format.
  Returns:
    bool: whether this is for a saved model conversion.
  """
  return input_format == common.TF_SAVED_MODEL

def available_output_formats(answers):
  """Generate the output formats for given input format.
  Args:
    ansowers: user selected parameter dict.
  """
  input_format = answers[common.INPUT_FORMAT]
  if input_format == common.KERAS_MODEL:
    return [{
        'key': 'g', # shortcut key for the option
        'name': 'Tensorflow.js Graph Model',
        'value': common.TFJS_GRAPH_MODEL,
    }, {
        'key': 'l',
        'name': 'TensoFlow.js Layers Model',
        'value': common.TFJS_LAYERS_MODEL,
    }]
  if input_format == common.TFJS_LAYERS_MODEL:
    return [{
        'key': 'k', # shortcut key for the option
        'name': 'Keras Model',
        'value': common.KERAS_MODEL,
    }, {
        'key': 's',
        'name': 'Keras Saved Model',
        'value': common.KERAS_SAVED_MODEL,
    }, {
        'key': 'l',
        'name': 'TensoFlow.js Layers Model',
        'value': common.TFJS_LAYERS_MODEL,
    }]
  return []


def available_tags(answers):
  """Generate the available saved model tags from the proto file.
  Args:
    ansowers: user selected parameter dict.
  """
  if is_saved_model(answers[common.INPUT_FORMAT]):
    saved_model = loader_impl.parse_saved_model(answers[common.INPUT_PATH])
    tags = []
    for meta_graph in saved_model.meta_graphs:
      tags.append(",".join(meta_graph.meta_info_def.tags))
    return tags
  return []


def available_signature_names(answers):
  """Generate the available saved model signatures from the proto file
    and selected tags.
  Args:
    ansowers: user selected parameter dict.
  """
  if (is_saved_model(answers[common.INPUT_FORMAT]) and
      common.SAVED_MODEL_TAGS in answers):
    path = answers[common.INPUT_PATH]
    tags = answers[common.SAVED_MODEL_TAGS]
    saved_model = loader_impl.parse_saved_model(path)
    for meta_graph in saved_model.meta_graphs:
      if tags == ",".join(meta_graph.meta_info_def.tags):
        signatures = []
        for key in meta_graph.signature_def:
          input_nodes = meta_graph.signature_def[key].inputs
          output_nodes = meta_graph.signature_def[key].outputs
          signatures.append(
              {'value': key,
               'name': format_signature(key, input_nodes, output_nodes)})
        return signatures
  return []


def format_signature(name, input_nodes, output_nodes):
  string = "signature name: %s\n" % name
  string += "        inputs: %s" % format_nodes(input_nodes)
  string += "        outputs: %s" % format_nodes(output_nodes)
  return string


def format_nodes(nodes):
  string = "%s of %s\n" % (3 if len(nodes) > 3 else len(nodes), len(nodes))
  count = 0
  for key in nodes:
    value = nodes[key]
    string += "              name: %s, " % value.name
    string += "dtype: %s, " % types_pb2.DataType.Name(value.dtype)
    if value.tensor_shape.unknown_rank:
      string += "shape: Unknown\n"
    else:
      string += "shape: %s\n" % [x.size for x in value.tensor_shape.dim]
    count += 1
    if count >= 3:
      break
  return string


def input_format_string(base, target_format, detected_format):
  if target_format == detected_format:
    return base + ' *'
  else:
    return base


def input_format_message(detected_input_format):
  message = 'What is your input model format? '
  if detected_input_format:
    message += '(auto-detected format is marked with *)'
  else:
    message += '(model format cannot be detected.) '
  return message

def update_output_path(output_path, params):
  output_path = os.path.expanduser(output_path.strip())
  if (common.OUTPUT_FORMAT in params and
      params[common.OUTPUT_FORMAT] == common.KERAS_MODEL):
    if os.path.isdir(output_path):
      output_path = os.path.join(output_path, 'model.H5')
  return output_path

def input_formats(detected_format):
  formats = [{
      'key': 'k',
      'name': input_format_string('Keras (HDF5)', common.KERAS_MODEL,
                                  detected_format),
      'value': common.KERAS_MODEL
  }, {
      'key': 'e',
      'name': input_format_string('Tensorflow Keras Saved Model',
                                  common.KERAS_SAVED_MODEL,
                                  detected_format),
      'value': common.KERAS_SAVED_MODEL,
  }, {
      'key': 's',
      'name': input_format_string('Tensorflow Saved Model',
                                  common.TF_SAVED_MODEL,
                                  detected_format),
      'value': common.TF_SAVED_MODEL,
  }, {
      'key': 'h',
      'name': input_format_string('TFHub Module',
                                  common.TF_HUB_MODEL,
                                  detected_format),
      'value': common.TF_HUB_MODEL,
  }, {
      'key': 'l',
      'name': input_format_string('TensoFlow.js Layers Model',
                                  common.TFJS_LAYERS_MODEL,
                                  detected_format),
      'value': common.TFJS_LAYERS_MODEL,
  }]
  formats.sort(key=lambda x: x['value'] != detected_format)
  return formats


def run(dryrun):
  print('Welcome to TensorFlow.js Converter.')
  input_path = [{
      'type': 'input',
      'name': common.INPUT_PATH,
      'message': 'Please provide the path of model file or '
                 'the directory that contains model files. \n'
                 'If you are converting TFHub module please provide the URL.',
      'filter': os.path.expanduser,
      'validate':
          lambda path: 'Please enter a valid path' if not path else True
  }]

  input_params = PyInquirer.prompt(input_path, style=prompt_style)
  detected_input_format, normalized_path = detect_input_format(
      input_params[common.INPUT_PATH])
  input_params[common.INPUT_PATH] = normalized_path

  formats = [
      {
          'type': 'list',
          'name': common.INPUT_FORMAT,
          'message': input_format_message(detected_input_format),
          'choices': input_formats(detected_input_format)
      }, {
          'type': 'list',
          'name': common.OUTPUT_FORMAT,
          'message': 'What is your output format?',
          'choices': available_output_formats,
          'when': lambda answers: value_in_list(answers, common.INPUT_FORMAT,
                                                (common.KERAS_MODEL,
                                                 common.TFJS_LAYERS_MODEL))
      }
  ]
  format_params = PyInquirer.prompt(formats, input_params, style=prompt_style)
  message = input_path_message(format_params)

  questions = [
      {
          'type': 'input',
          'name': common.INPUT_PATH,
          'message': message,
          'filter': expand_input_path,
          'validate': lambda value: validate_input_path(
              value, format_params[common.INPUT_FORMAT]),
          'when': lambda answers: (not detected_input_format)
      },
      {
          'type': 'list',
          'name': common.SAVED_MODEL_TAGS,
          'choices': available_tags,
          'message': 'What is tags for the saved model?',
          'when': lambda answers: (is_saved_model(answers[common.INPUT_FORMAT])
                                   and
                                   (common.OUTPUT_FORMAT not in format_params
                                    or format_params[common.OUTPUT_FORMAT] ==
                                    common.TFJS_GRAPH_MODEL))
      },
      {
          'type': 'list',
          'name': common.SIGNATURE_NAME,
          'message': 'What is signature name of the model?',
          'choices': available_signature_names,
          'when': lambda answers: (is_saved_model(answers[common.INPUT_FORMAT])
                                   and
                                   (common.OUTPUT_FORMAT not in format_params
                                    or format_params[common.OUTPUT_FORMAT] ==
                                    common.TFJS_GRAPH_MODEL))
      },
      {
          'type': 'list',
          'name': 'quantize',
          'message': 'Do you want to compress the model? '
                     '(this will decrease the model precision.)',
          'choices': [{
              'name': 'No compression (Higher accuracy)',
              'value': None
          }, {
              'name': 'float16 quantization '
                      '(2x smaller, Minimal accuracy loss)',
              'value': 'float16'
          }, {
              'name': 'uint16 affine quantization (2x smaller, Accuracy loss)',
              'value': 'uint16'
          }, {
              'name': 'uint8 affine quantization (4x smaller, Accuracy loss)',
              'value': 'uint8'
          }]
      },
      {
          'type': 'input',
          'name': common.QUANTIZATION_TYPE_FLOAT16,
          'message': 'Please enter the layers to apply float16 quantization '
                     '(2x smaller, minimal accuracy tradeoff).\n'
                     'Supports wildcard expansion with *, e.g., conv/*/weights',
          'default': '*',
          'when': lambda answers:
                  value_in_list(answers, 'quantize', ('float16'))
      },
      {
          'type': 'input',
          'name': common.QUANTIZATION_TYPE_UINT8,
          'message': 'Please enter the layers to apply affine 1-byte integer '
                     'quantization (4x smaller, accuracy tradeoff).\n'
                     'Supports wildcard expansion with *, e.g., conv/*/weights',
          'default': '*',
          'when': lambda answers:
                  value_in_list(answers, 'quantize', ('uint8'))
      },
      {
          'type': 'input',
          'name': common.QUANTIZATION_TYPE_UINT16,
          'message': 'Please enter the layers to apply affine 2-byte integer '
                     'quantization (2x smaller, accuracy tradeoff).\n'
                     'Supports wildcard expansion with *, e.g., conv/*/weights',
          'default': '*',
          'when': lambda answers:
                  value_in_list(answers, 'quantize', ('uint16'))
      },
      {
          'type': 'input',
          'name': common.WEIGHT_SHARD_SIZE_BYTES,
          'message': 'Please enter shard size (in bytes) of the weight files?',
          'default': str(4 * 1024 * 1024),
          'validate':
              lambda size: ('Please enter a positive integer' if not
                            (size.isdigit() and int(size) > 0) else True),
          'when': lambda answers: (value_in_list(answers, common.OUTPUT_FORMAT,
                                                 (common.TFJS_LAYERS_MODEL,
                                                  common.TFJS_GRAPH_MODEL)) or
                                   value_in_list(answers, common.INPUT_FORMAT,
                                                 (common.TF_SAVED_MODEL,
                                                  common.TF_HUB_MODEL)))
      },
      {
          'type': 'confirm',
          'name': common.SPLIT_WEIGHTS_BY_LAYER,
          'message': 'Do you want to split weights by layers?',
          'default': False,
          'when': lambda answers: (value_in_list(answers, common.OUTPUT_FORMAT,
                                                 (common.TFJS_LAYERS_MODEL)) and
                                   value_in_list(answers, common.INPUT_FORMAT,
                                                 (common.KERAS_MODEL,
                                                  common.KERAS_SAVED_MODEL)))
      },
      {
          'type': 'confirm',
          'name': common.SKIP_OP_CHECK,
          'message': 'Do you want to skip op validation? \n'
                     'This will allow conversion of unsupported ops, \n'
                     'you can implement them as custom ops in tfjs-converter.',
          'default': False,
          'when': lambda answers: value_in_list(answers, common.INPUT_FORMAT,
                                                (common.TF_SAVED_MODEL,
                                                 common.TF_HUB_MODEL))
      },
      {
          'type': 'confirm',
          'name': common.STRIP_DEBUG_OPS,
          'message': 'Do you want to strip debug ops? \n'
                     'This will improve model execution performance.',
          'default': True,
          'when': lambda answers: value_in_list(answers, common.INPUT_FORMAT,
                                                (common.TF_SAVED_MODEL,
                                                 common.TF_HUB_MODEL))
      },
      {
          'type': 'confirm',
          'name': common.CONTROL_FLOW_V2,
          'message': 'Do you want to enable Control Flow V2 ops? \n'
                     'This will improve branch and loop execution performance.',
          'default': True,
          'when': lambda answers: value_in_list(answers, common.INPUT_FORMAT,
                                                (common.TF_SAVED_MODEL,
                                                 common.TF_HUB_MODEL))
      },
      {
          'type': 'input',
          'name': common.METADATA,
          'message': 'Do you want to provide metadata? \n'
                     'Provide your own metadata in the form: \n'
                     'metadata_key:path/metadata.json \n'
                     'Separate multiple metadata by comma.'
      }
  ]
  params = PyInquirer.prompt(questions, format_params, style=prompt_style)

  output_options = [
      {
          'type': 'input',
          'name': common.OUTPUT_PATH,
          'message': 'Which directory do you want to save '
                     'the converted model in?',
          'filter': lambda path: update_output_path(path, params),
          'validate': lambda path: len(path) > 0
      },
      {
          'type': 'confirm',
          'message': 'The output already directory exists, '
                     'do you want to overwrite it?',
          'name': 'overwrite_output_path',
          'default': False,
          'when': lambda ans: output_path_exists(ans[common.OUTPUT_PATH])
      }
  ]

  while (common.OUTPUT_PATH not in params or
         output_path_exists(params[common.OUTPUT_PATH]) and
         not params['overwrite_output_path']):
    params = PyInquirer.prompt(output_options, params, style=prompt_style)

  arguments = generate_arguments(params)
  print('converter command generated:')
  print('tensorflowjs_converter %s' % ' '.join(arguments))
  print('\n\n')

  log_file = os.path.join(tempfile.gettempdir(), 'converter_error.log')
  if not dryrun:
    try:
      converter.convert(arguments)
      print('\n\nFile(s) generated by conversion:')

      print("Filename {0:25} Size(bytes)".format(''))
      total_size = 0
      output_path = params[common.OUTPUT_PATH]
      if os.path.isfile(output_path):
        output_path = os.path.dirname(output_path)
      for basename in sorted(os.listdir(output_path)):
        filename = os.path.join(output_path, basename)
        size = os.path.getsize(filename)
        print("{0:35} {1}".format(basename, size))
        total_size += size
      print("Total size:{0:24} {1}".format('', total_size))
    except BaseException:
      exc_type, exc_value, exc_traceback = sys.exc_info()
      lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
      with open(log_file, 'a') as writer:
        writer.write(''.join(line for line in lines))
      print('Conversion failed, please check error log file %s.' % log_file)

def pip_main():
  """Entry point for pip-packaged binary.

  Note that pip-packaged binary calls the entry method without
  any arguments, which is why this method is needed in addition to the
  `main` method below.
  """
  main([' '.join(sys.argv[1:])])


def main(argv):
  if argv[0] and not argv[0] == '--dryrun':
    print("Usage: tensorflowjs_wizard [--dryrun]")
    sys.exit(1)
  dry_run = argv[0] == '--dryrun'
  run(dry_run)

if __name__ == '__main__':
  tf.app.run(main=main, argv=[' '.join(sys.argv[1:])])
